Blame = {
  State =  { playerId = ""},
  Action = { },
  blameOnNextHit = 0,
  SavedIds = {},
  buffs = {},
}

-- saved var defaults
Blame_queueDepth = 10

Blame.State.playerId = Inspect.Unit.Lookup("player")

local function generateHitLine(event, timedelta)
  sb = {}
  if (timedelta < 120) then
    table.insert(sb, "-(")
    table.insert(sb, string.format("%.1f", timedelta))
    table.insert(sb, ")")
  else
    table.insert(sb, "---");
  end
  table.insert(sb, event.caster)
  if (event.over > 0) then
    table.insert(sb, " killed you with a hit for ")
  else
    table.insert(sb, " hit you for ")
  end
  table.insert(sb, event.amt)
  if (event.over > 0) then
    table.insert(sb, " (" .. event.over .. " overkill)")
  end
  if (event.crit) then
    table.insert(sb, " (crit)");
  end
  table.insert(sb, " with ");
  table.insert(sb, event.ability)
  if (event.absorbed > 0 or event.blocked > 0 or event.intercept > 0 or event.modified > 0) then
    table.insert(sb, " |");
    if (event.absorbed > 0) then
      table.insert(sb, "absorb " .. event.absorbed .. "|")
    end
    if (event.blocked > 0) then
      table.insert(sb, "block " .. event.blocked .. "|")
    end
    if (event.intercept > 0) then
      table.insert(sb, "interc " .. event.intercept .. "|")
    end
    if (event.modified > 0) then
      table.insert(sb, "modified " .. event.modified .. "|")
    end
  end
  return table.concat(sb)
end

local function generateHealLine(event, timedelta)
  sb = {}
  if (timedelta < 120) then
    table.insert(sb, "+(")
    table.insert(sb, string.format("%.1f", timedelta))
    table.insert(sb, ")")
  else
    table.insert(sb, "---");
  end
  table.insert(sb, event.caster)
  table.insert(sb, " healed you for ")
  table.insert(sb, event.amt)
  if (event.over > 0) then
    table.insert(sb, "(")
    table.insert(sb, event.over)
    table.insert(sb, " overheal)")
  end
  if (event.crit) then
    table.insert(sb, " (crit)")
  end
  table.insert(sb, " with ")
  table.insert(sb, event.ability)
  return table.concat(sb);
end

local function printDebuffs()
  if (Blame.buffs == nil) then return end
  print("Debuffs:")
  local details = Blame.buffs
  for trash, detail in pairs(details) do
    if (detail) then
      if (detail.debuff ~= nil) then
        local sb = {}
        table.insert(sb, detail.name)
        if (detail.stack ~= nil) then
          table.insert(sb, "(")
          table.insert(sb, detail.stack)
          table.insert(sb, ")")
        end
        print(table.concat(sb))
      end
    end
  end
end

local function printBlame(detail, overkill)
  sb = {}
  table.insert(sb, "You were killed by ")
  table.insert(sb, detail.name)
  if (overkill ~= nil and overkill > 0) then
    table.insert(sb, " (")
    table.insert(sb, overkill)
    table.insert(sb, " overkill) ")    
  end
  table.insert(sb, ".  Here are the last few events:")
  print(table.concat(sb))
  for i=1,#Blame.Action do
    timedelta = Blame.Action[#Blame.Action].time - Blame.Action[i].time
    if (Blame.Action[i]["type"] == "hit") then
      print(generateHitLine(Blame.Action[i], timedelta))
    else
      print(generateHealLine(Blame.Action[i], timedelta))
    end
  end
  printDebuffs()
end


local function trackBlameHit(params)
  if Blame_queueDepth == 0 then return end
  if params.target ~= Blame.State.playerId then return end
  time = Inspect.Time.Real()
  if #Blame.Action > 1 then
    if ((time - Blame.Action[#Blame.Action].time) > 10) then
      Blame.SavedIds = {}
    end
  end
  detail = Inspect.Unit.Detail(params.caster)
  -- this can happen if the caster of the hit is already dead
  if (detail == nil) then
    detail = Blame.SavedIds[params.caster]
    if detail == nil then
      detail = { name = "unknown" }
    end
  else 
    Blame.SavedIds[params.caster] = detail
  end
  local entry = {}
  entry["type"] = "hit"
  entry["caster"] = detail.name
  entry["amt"] = params.damage or 0
  entry["crit"] = params.crit
  entry["over"] = params.overkill or 0
  entry["absorbed"] = params.damageAbsorbed or 0
  entry["blocked"] = params.damageBlocked or 0
  entry["intercept"] = params.damageIntercepted or 0
  entry["modified"] = params.damageModified or 0
  entry["ability"] = params.abilityName
  entry["time"] = time
  table.insert(Blame.Action, entry)
  if (#Blame.Action > Blame_queueDepth) then
    table.remove(Blame.Action, 1)
  end
  if Blame.blameOnNextHit == 1 then
    printBlame(detail, params.overkill)
    Blame.blameOnNextHit = 0
  end
end

local function trackBlameHeal(params)
  time = Inspect.Time.Real()
  if Blame_queueDepth == 0 then return end
  if params.target ~= Blame.State.playerId then return end
  if not params.heal then return end
  local stringBuilder = {}
  detail = Inspect.Unit.Detail(params.caster)
  if (detail == nil) then return end
  local entry = {}
  entry["type"] = "heal"
  entry["caster"] = detail.name
  entry["amt"] = params.heal
  entry["crit"] = params.crit
  entry["over"] = params.overheal or 0
  entry["ability"] = params.abilityName
  entry["time"] = time
  table.insert(Blame.Action, entry)
  if (#Blame.Action > Blame_queueDepth) then
    table.remove(Blame.Action, 1)
  end
end

local function trackDeath(params)
  if Blame_queueDepth == 0 then return end
  if params.target ~= Blame.State.playerId then return end
  Blame.blameOnNextHit = 1
end

local function trackBuffChange(unit, bufflist)
  if (Blame.blameOnNextHit == 1) then return end
  if (unit ~= Blame.State.playerId) then return end
  if (Inspect.Unit.Detail(Blame.State.playerId).health > 0) then
    Blame.buffs = Inspect.Buff.Detail("player", Inspect.Buff.List("player"))
  end
end

local function onDemandBlame()
  if (#Blame.Action == 0) then
    print("No activity to report.")
    return
  end
  sb = {}
  table.insert(sb, "Here are the last ")
  table.insert(sb, #Blame.Action)
  table.insert(sb, " combat events:")        
  print(table.concat(sb))
  for i=1,#Blame.Action do
    timedelta = Blame.Action[#Blame.Action].time - Blame.Action[i].time
    if (Blame.Action[i]["type"] == "hit") then
      print(generateHitLine(Blame.Action[i], timedelta))
    else
      print(generateHealLine(Blame.Action[i], timedelta))
    end
  end
  printDebuffs()
end

local function setDepth(line)
  newMax = tonumber(line)
  if newMax then
    if newMax < 100 and newMax > 3 then
      Blame_queueDepth = newMax
      print("Blame queue set to " .. Blame_queueDepth)
    elseif newMax == 0 then
      Blame_queueDepth = 0
      Blame.Action = {}
      print("Blame off")
    else
      print("Depth must be a number between 3 and 100 or 0 to turn Blame off");
    end
  else
    print("Usage:  /blameconf <depth> where <depth> must be a number between 3 and 100 or 0 to turn Blame off")
    if (Blame_queueDepth == 0) then
      print("Currently Blame is off.");
    else
      print("Current Blame depth is " .. Blame_queueDepth)
    end
  end
end

local function onInit(addon)
  if addon == "Blame" then
    print("Started.")
    print("/blame to view log immediately")
    print("/blameconf to configure blame")
  end
end

table.insert(Event.Addon.Load.End, {onInit, "Blame", "Setup"})
table.insert(Event.Combat.Damage, {trackBlameHit, "Blame", "aCombatHitEvent"})
table.insert(Event.Combat.Heal, {trackBlameHeal, "Blame", "aCombatHealEvent"})
table.insert(Event.Combat.Death, {trackDeath, "Blame", "aDeathEvent"})
table.insert(Event.Buff.Add, {trackBuffChange, "Blame", "aBuffAddEvent"})
table.insert(Event.Buff.Change, {trackBuffChange, "Blame", "aBuffChangeEvent"})
table.insert(Event.Buff.Remove, {trackBuffChange, "Blame", "aBuffRemoveEvent"})
table.insert(Command.Slash.Register("blame"), {onDemandBlame, "Blame", "Show Blame"})
table.insert(Command.Slash.Register("blameconf"), {setDepth, "Blame", "Config Blame"})